@@ -79,7 +79,7 @@
 #  define INTERFACES_KEY       "Interfaces"
 #  define DOMAIN_KEY           "Domain"
 #  define DHCPDOMAIN_KEY       "DhcpDomain"
-
+#  define PATH_RESOLV_CONF     ""
 #elif defined(WATT32)
 
 #  define PATH_RESOLV_CONF "/dev/ENV/etc/resolv.conf"
@@ -92,7 +92,8 @@ W32_FUNC const char *_w32_GetHostsFile(void);
 
 #elif defined(__riscos__)
 
-#  define PATH_HOSTS "InetDBase:Hosts"
+#  define PATH_RESOLV_CONF ""
+#  define PATH_HOSTS       "InetDBase:Hosts"
 
 #elif defined(__HAIKU__)
 
@@ -117,10 +118,11 @@ typedef struct ares_rand_state ares_rand_state;
 
 #include "ares__llist.h"
 #include "ares__slist.h"
-#include "ares__htable_stvp.h"
+#include "ares__htable_strvp.h"
+#include "ares__htable_szvp.h"
 #include "ares__htable_asvp.h"
 #include "ares__buf.h"
-#include "ares_dns_record.h"
+#include "ares_dns_private.h"
 
 #ifndef HAVE_GETENV
 #  include "ares_getenv.h"
@@ -149,20 +151,6 @@ typedef struct ares_rand_state ares_rand_state;
 
 /********* EDNS defines section ******/
 
-struct ares_addr {
-  int family;
-
-  union {
-    struct in_addr       addr4;
-    struct ares_in6_addr addr6;
-  } addr;
-
-  unsigned short udp_port; /* stored in network order */
-  unsigned short tcp_port; /* stored in network order */
-};
-
-#define addrV4 addr.addr4
-#define addrV6 addr.addr6
 
 struct query;
 
@@ -179,8 +167,13 @@ struct server_connection {
 };
 
 struct server_state {
-  size_t                    idx; /* index for server in ares_channel */
+  size_t                    idx; /* index for server in system configuration */
+  size_t                    consec_failures; /* Consecutive query failure count
+                                              * can be hard errors or timeouts
+                                              */
   struct ares_addr          addr;
+  unsigned short            udp_port;
+  unsigned short            tcp_port;
 
   ares__llist_t            *connections;
   struct server_connection *tcp_conn;
@@ -192,61 +185,44 @@ struct server_state {
   /* TCP output queue */
   ares__buf_t              *tcp_send;
 
-  /* Which incarnation of this connection is this? We don't want to
-   * retransmit requests into the very same socket, but if the server
-   * closes on us and we re-open the connection, then we do want to
-   * re-send. */
-  size_t                    tcp_connection_generation;
-
   /* Link back to owning channel */
-  ares_channel              channel;
+  ares_channel_t           *channel;
 };
 
 /* State to represent a DNS query */
 struct query {
   /* Query ID from qbuf, for faster lookup, and current timeout */
-  unsigned short                  qid; /* host byte order */
-  struct timeval                  timeout;
-  ares_channel                    channel;
+  unsigned short            qid; /* host byte order */
+  struct timeval            timeout;
+  ares_channel_t           *channel;
 
   /*
    * Node object for each list entry the query belongs to in order to
    * make removal operations O(1).
    */
-  ares__slist_node_t             *node_queries_by_timeout;
-  ares__llist_node_t             *node_queries_to_conn;
-  ares__llist_node_t             *node_all_queries;
+  ares__slist_node_t       *node_queries_by_timeout;
+  ares__llist_node_t       *node_queries_to_conn;
+  ares__llist_node_t       *node_all_queries;
 
-  /* connection handle for validation purposes */
-  const struct server_connection *conn;
+  /* connection handle query is associated with */
+  struct server_connection *conn;
 
-  /* Query buf with length at beginning, for TCP transmission */
-  unsigned char                  *tcpbuf;
-  size_t                          tcplen;
+  /* Arguments passed to ares_send() */
+  unsigned char            *qbuf;
+  size_t                    qlen;
 
-  /* Arguments passed to ares_send() (qbuf points into tcpbuf) */
-  const unsigned char            *qbuf;
-  size_t                          qlen;
-  ares_callback                   callback;
-  void                           *arg;
+  ares_callback             callback;
+  void                     *arg;
 
   /* Query status */
-  size_t try_count; /* Number of times we tried this query already. */
-  size_t server;    /* Server this query has last been sent to. */
-  struct query_server_info *server_info; /* per-server state */
-  ares_bool_t               using_tcp;
-  ares_status_t             error_status;
-  size_t      timeouts;   /* number of timeouts we saw for this request */
+  size_t        try_count; /* Number of times we tried this query already. */
+  ares_bool_t   using_tcp;
+  ares_status_t error_status;
+  size_t        timeouts; /* number of timeouts we saw for this request */
   ares_bool_t no_retries; /* do not perform any additional retries, this is set
                            * when a query is to be canceled */
 };
 
-/* Per-server state for a query */
-struct query_server_info {
-  ares_bool_t skip_server; /* should we skip server, due to errors, etc? */
-  size_t tcp_connection_generation; /* into which TCP connection did we send? */
-};
-
 /* An IP address pattern; matches an IP address X if X & mask == addr */
 #define PATTERN_MASK 0x1
 #define PATTERN_CIDR 0x2
@@ -267,12 +243,16 @@ struct apattern {
   unsigned short type;
 };
 
+struct ares_hosts_file;
+typedef struct ares_hosts_file ares_hosts_file_t;
+
 struct ares_channeldata {
   /* Configuration data */
   unsigned int         flags;
   size_t               timeout; /* in milliseconds */
   size_t               tries;
   size_t               ndots;
+  size_t               maxtimeout;                 /* in milliseconds */
   ares_bool_t          rotate;
   unsigned short       udp_port;                   /* stored in network order */
   unsigned short       tcp_port;                   /* stored in network order */
@@ -293,23 +273,18 @@ struct ares_channeldata {
   unsigned int         local_ip4;
   unsigned char        local_ip6[16];
 
-  /* Server addresses and communications state */
-  struct server_state *servers;
-  size_t               nservers;
+  /* Server addresses and communications state. Sorted by least consecutive
+   * failures, followed by the configuration order if failures are equal. */
+  ares__slist_t       *servers;
 
-  /* random state to use when generating new ids */
+  /* random state to use when generating new ids and generating retry penalties
+   */
   ares_rand_state     *rand_state;
 
-  /* Generation number to use for the next TCP socket open/close */
-  size_t               tcp_connection_generation;
-
-  /* Last server we sent a query to. */
-  size_t               last_server;
-
   /* All active queries in a single list */
   ares__llist_t       *all_queries;
   /* Queries bucketed by qid, for quickly dispatching DNS responses: */
-  ares__htable_stvp_t *queries_by_qid;
+  ares__htable_szvp_t *queries_by_qid;
 
   /* Queries bucketed by timeout, for quickly handling timeouts: */
   ares__slist_t       *queries_by_timeout;
@@ -340,6 +315,9 @@ struct ares_channeldata {
 
   /* Maximum UDP queries per connection allowed */
   size_t                              udp_max_queries;
+
+  /* Cache of local hosts file */
+  ares_hosts_file_t                  *hf;
 };
 
 /* Does the domain end in ".onion" or ".onion."? Case-insensitive. */
@@ -353,26 +331,28 @@ void         *ares_malloc_zero(size_t size);
 void         *ares_realloc_zero(void *ptr, size_t orig_size, size_t new_size);
 
 /* return true if now is exactly check time or later */
-ares_bool_t   ares__timedout(struct timeval *now, struct timeval *check);
+ares_bool_t   ares__timedout(const struct timeval *now,
+                             const struct timeval *check);
 
 /* Returns one of the normal ares status codes like ARES_SUCCESS */
-ares_status_t ares__send_query(ares_channel channel, struct query *query,
-                               struct timeval *now);
+ares_status_t ares__send_query(struct query *query, struct timeval *now);
+ares_status_t ares__requeue_query(struct query *query, struct timeval *now);
 
 /* Identical to ares_query, but returns a normal ares return code like
  * ARES_SUCCESS, and can be passed the qid by reference which will be
  * filled in on ARES_SUCCESS */
-ares_status_t ares_query_qid(ares_channel channel, const char *name,
+ares_status_t ares_query_qid(ares_channel_t *channel, const char *name,
                              int dnsclass, int type, ares_callback callback,
                              void *arg, unsigned short *qid);
 /* Identical to ares_send() except returns normal ares return codes like
  * ARES_SUCCESS */
-ares_status_t ares_send_ex(ares_channel channel, const unsigned char *qbuf,
-                           size_t qlen, ares_callback callback, void *arg);
+ares_status_t ares_send_ex(ares_channel_t *channel, const unsigned char *qbuf,
+                           size_t qlen, ares_callback callback, void *arg,
+                           unsigned short *qid);
 void          ares__close_connection(struct server_connection *conn);
 void          ares__close_sockets(struct server_state *server);
-void          ares__check_cleanup_conn(ares_channel channel, ares_socket_t fd);
-ares_status_t ares__get_hostent(FILE *fp, int family, struct hostent **host);
+void          ares__check_cleanup_conn(const ares_channel_t     *channel,
+                                       struct server_connection *conn);
 ares_status_t ares__read_line(FILE *fp, char **buf, size_t *bufsize);
 void          ares__free_query(struct query *query);
 
@@ -394,19 +374,41 @@ ares_status_t  ares__expand_name_for_response(const unsigned char *encoded,
 ares_status_t  ares_expand_string_ex(const unsigned char *encoded,
                                      const unsigned char *abuf, size_t alen,
                                      unsigned char **s, size_t *enclen);
-ares_status_t  ares__init_servers_state(ares_channel channel);
-void           ares__destroy_servers_state(ares_channel channel);
-ares_status_t  ares__single_domain(ares_channel channel, const char *name,
-                                   char **s);
-ares_status_t  ares__cat_domain(const char *name, const char *domain, char **s);
-ares_status_t  ares__sortaddrinfo(ares_channel               channel,
-                                  struct ares_addrinfo_node *ai_node);
-ares_status_t  ares__readaddrinfo(FILE *fp, const char *name,
-                                  unsigned short                    port,
-                                  const struct ares_addrinfo_hints *hints,
-                                  struct ares_addrinfo             *ai);
+ares_status_t  ares__init_servers_state(ares_channel_t *channel);
+ares_status_t  ares__init_by_options(ares_channel_t            *channel,
+                                     const struct ares_options *options,
+                                     int                        optmask);
+ares_status_t  ares__init_by_sysconfig(ares_channel_t *channel);
 
-void           ares__freeaddrinfo_nodes(struct ares_addrinfo_node *ai_node);
+typedef struct {
+  ares__llist_t   *sconfig;
+  struct apattern *sortlist;
+  size_t           nsortlist;
+  char           **domains;
+  size_t           ndomains;
+  char            *lookups;
+  size_t           ndots;
+  size_t           tries;
+  ares_bool_t      rotate;
+  size_t           timeout_ms;
+} ares_sysconfig_t;
+
+ares_status_t ares__init_by_environment(ares_sysconfig_t *sysconfig);
+
+ares_status_t ares__init_sysconfig_files(const ares_channel_t *channel,
+                                         ares_sysconfig_t     *sysconfig);
+ares_status_t ares__parse_sortlist(struct apattern **sortlist, size_t *nsort,
+                                   const char *str);
+
+void          ares__destroy_servers_state(ares_channel_t *channel);
+ares_status_t ares__single_domain(const ares_channel_t *channel,
+                                  const char *name, char **s);
+ares_status_t ares__cat_domain(const char *name, const char *domain, char **s);
+ares_status_t ares__sortaddrinfo(ares_channel_t            *channel,
+                                 struct ares_addrinfo_node *ai_node);
+
+void          ares__freeaddrinfo_nodes(struct ares_addrinfo_node *ai_node);
+ares_bool_t   ares__is_localhost(const char *name);
 
 struct ares_addrinfo_node    *
   ares__append_addrinfo_node(struct ares_addrinfo_node **ai_node);
@@ -440,12 +442,96 @@ ares_status_t ares__addrinfo2addrttl(const struct ares_addrinfo *ai, int family,
 ares_status_t ares__addrinfo_localhost(const char *name, unsigned short port,
                                        const struct ares_addrinfo_hints *hints,
                                        struct ares_addrinfo             *ai);
-
-ares_socket_t ares__open_socket(ares_channel channel, int af, int type,
+ares_status_t ares__open_connection(ares_channel_t      *channel,
+                                    struct server_state *server,
+                                    ares_bool_t          is_tcp);
+ares_socket_t ares__open_socket(ares_channel_t *channel, int af, int type,
                                 int protocol);
+ares_ssize_t  ares__socket_write(ares_channel_t *channel, ares_socket_t s,
+                                 const void *data, size_t len);
+ares_ssize_t  ares__socket_recvfrom(ares_channel_t *channel, ares_socket_t s,
+                                    void *data, size_t data_len, int flags,
+                                    struct sockaddr *from,
+                                    ares_socklen_t  *from_len);
+ares_ssize_t  ares__socket_recv(ares_channel_t *channel, ares_socket_t s,
+                                void *data, size_t data_len);
 void          ares__close_socket(ares_channel, ares_socket_t);
-int           ares__connect_socket(ares_channel channel, ares_socket_t sockfd,
-                                   const struct sockaddr *addr, ares_socklen_t addrlen);
+int         ares__connect_socket(ares_channel_t *channel, ares_socket_t sockfd,
+                                 const struct sockaddr *addr, ares_socklen_t addrlen);
+ares_bool_t ares__is_hostnamech(int ch);
+void        ares__destroy_server(struct server_state *server);
+
+ares_status_t ares__servers_update(ares_channel_t *channel,
+                                   ares__llist_t  *server_list,
+                                   ares_bool_t     user_specified);
+ares_status_t ares__sconfig_append(ares__llist_t         **sconfig,
+                                   const struct ares_addr *addr,
+                                   unsigned short          udp_port,
+                                   unsigned short          tcp_port);
+ares_status_t ares__sconfig_append_fromstr(ares__llist_t **sconfig,
+                                           const char     *str);
+ares__llist_t *
+  ares_in_addr_to_server_config_llist(const struct in_addr *servers,
+                                      size_t                nservers);
+
+struct ares_hosts_entry;
+typedef struct ares_hosts_entry ares_hosts_entry_t;
+
+void                            ares__hosts_file_destroy(ares_hosts_file_t *hf);
+ares_status_t ares__hosts_search_ipaddr(ares_channel_t *channel,
+                                        ares_bool_t use_env, const char *ipaddr,
+                                        const ares_hosts_entry_t **entry);
+ares_status_t ares__hosts_search_host(ares_channel_t *channel,
+                                      ares_bool_t use_env, const char *host,
+                                      const ares_hosts_entry_t **entry);
+ares_status_t ares__hosts_entry_to_hostent(const ares_hosts_entry_t *entry,
+                                           int                       family,
+                                           struct hostent          **hostent);
+ares_status_t ares__hosts_entry_to_addrinfo(const ares_hosts_entry_t *entry,
+                                            const char *name, int family,
+                                            unsigned short        port,
+                                            ares_bool_t           want_cnames,
+                                            struct ares_addrinfo *ai);
+ares_bool_t   ares__isprint(int ch);
+
+
+/*! Parse a compressed DNS name as defined in RFC1035 starting at the current
+ *  offset within the buffer.
+ *
+ *  It is assumed that either a const buffer is being used, or before
+ *  the message processing was started that ares__buf_reclaim() was called.
+ *
+ *  \param[in]  buf        Initialized buffer object
+ *  \param[out] name       Pointer passed by reference to be filled in with
+ *                         allocated string of the parsed name that must be
+ *                         ares_free()'d by the caller.
+ *  \param[in] is_hostname if ARES_TRUE, will validate the character set for
+ *                         a valid hostname or will return error.
+ *  \return ARES_SUCCESS on success
+ */
+ares_status_t ares__dns_name_parse(ares__buf_t *buf, char **name,
+                                   ares_bool_t is_hostname);
+
+/*! Write the DNS name to the buffer in the DNS domain-name syntax as a
+ *  series of labels.  The maximum domain name length is 255 characters with
+ *  each label being a maximum of 63 characters.  If the validate_hostname
+ *  flag is set, it will strictly validate the character set.
+ *
+ *  \param[in,out]  buf   Initialized buffer object to write name to
+ *  \param[in,out]  list  Pointer passed by reference to maintain a list of
+ *                        domain name to indexes used for name compression.
+ *                        Pass NULL (not by reference) if name compression isn't
+ *                        desired.  Otherwise the list will be automatically
+ *                        created upon first entry.
+ *  \param[in]      validate_hostname Validate the hostname character set.
+ *  \param[in]      name              Name to write out, it may have escape
+ *                                    sequences.
+ *  \return ARES_SUCCESS on success, most likely ARES_EBADNAME if the name is
+ *          bad.
+ */
+ares_status_t ares__dns_name_write(ares__buf_t *buf, ares__llist_t **list,
+                                   ares_bool_t validate_hostname,
+                                   const char *name);
 
 #define ARES_SWAP_BYTE(a, b)           \
   do {                                 \
@@ -461,8 +547,15 @@ int           ares__connect_socket(ares_channel channel, ares_socket_t sockfd,
     }                                                             \
   } while (0)
 
+#define ARES_CONFIG_CHECK(x)                                             \
+  (x && x->lookups && ares__slist_len(x->servers) > 0 && x->ndots > 0 && \
+   x->timeout > 0 && x->tries > 0)
 
 size_t ares__round_up_pow2(size_t n);
 size_t ares__log2(size_t n);
+size_t ares__pow(size_t x, size_t y);
+size_t ares__count_digits(size_t n);
+size_t ares__count_hexdigits(size_t n);
+
 
 #endif /* __ARES_PRIVATE_H */
