@@ -84,6 +84,9 @@
 #  define HAVE_WS2TCPIP_H 1
 #endif
 
+#define HAVE_SYS_TYPES_H 1
+#define HAVE_SYS_STAT_H  1
+
 /* ---------------------------------------------------------------- */
 /*                        OTHER HEADER INFO                         */
 /* ---------------------------------------------------------------- */
